/*
 * timer.c - Event timers
 * 
 * Written 2002 by Werner Almesberger
 */


#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <math.h>
#include <signal.h>
#include <sys/time.h>

#include <pthread.h>


extern int hz(void);
extern int got_work(void);
extern int other_runnable(void);


#include "timer.h"


#define __TRACE(s) \
  write(2,s,sizeof(s));
#define TRACE(s) __TRACE(s "\r\n");

#undef TRACE
#define TRACE(s)


typedef struct _timer {
    struct timeval due;
    void (*fn)(void *data);
    void *data;
    struct _timer *next;
} TIMER;


static const int interactive = 0;

struct timeval now;

static TIMER *timers = NULL;
static double jiffie;
static struct timeval jiffie_tv;


static struct timeval jiffies2tv(unsigned long jiffies)
{
    struct timeval tv;
    double s;

    s = jiffies*jiffie;
    tv.tv_sec = floor(s);
    tv.tv_usec = (s-floor(s))*1e6;
    return tv;
}


unsigned long tv2jiffies(struct timeval tv)
{
    return floor(((double) tv.tv_sec+(double) tv.tv_usec/1e6)*(double) hz());
}


static struct timeval tv_add(struct timeval a,struct timeval b)
{
    a.tv_sec += b.tv_sec;
    a.tv_usec += b.tv_usec;
    while (a.tv_usec > 1000000) {
	a.tv_usec -= 1000000;
	a.tv_sec++;
    }
    return a;
}


void wake_kernel(unsigned delta_jiffies)
{
    TRACE("wake_kernel");
    set_timer(jiffies2tv(delta_jiffies),NULL,NULL);
}


static inline int tv_lt(const struct timeval a,const struct timeval b)
{
    if (a.tv_sec < b.tv_sec) return 1;
    if (a.tv_sec > b.tv_sec) return 0;
    return a.tv_usec < b.tv_usec;
}


static struct timeval last_tick = { 0,0 };
extern int use_sigalrm;

static void do_tick(void)
{
    raise(use_sigalrm ? SIGALRM : SIGVTALRM);
    last_tick = now;
}

static int timer_ready = 0;

static void have_timers(void)
{
    TIMER *curr;
    struct timeval next_tick;

    next_tick = tv_add(last_tick,jiffie_tv);
    while (tv_lt(next_tick,timers->due)) {
	now = next_tick;
	if (got_work()) {
	    do_tick();
	    return;
	}
	jiffies++;
	last_tick = now;
	next_tick = tv_add(now,jiffie_tv);
    }
    curr = timers;
    timers = timers->next;
    if (!curr->fn) timer_ready = 1;
    else {
	now = curr->due;
	curr->fn(curr->data);
    }
    free(curr);
}


static void no_timers(void)
{
    struct timeval next_tick;

    next_tick = tv_add(last_tick,jiffie_tv);
    if (interactive || got_work() || timer_ready) {
	timer_ready = 0;
	now = next_tick;
	do_tick();
	return;
    }
    if (other_runnable()) return;
    __TRACE("THE SYSTEM SEEMS IDLE, WHICH IS IMPOSSIBLE.\r\n");
}


void wait_kernel(void)
{
    TRACE("wait_kernel");
    if (timers) have_timers();
    else no_timers();
}


void set_timer(struct timeval tv,void (*fn)(void *data),void *data)
{
    TIMER *new,**walk;

    TRACE("set_timer");
    new = malloc(sizeof(TIMER));
    if (!new) abort();
    new->due = tv_add(now,tv);
    new->fn = fn;
    new->data = data;
    for (walk = &timers; *walk; walk = &(*walk)->next)
	if (!tv_lt((*walk)->due,new->due)) break;
    new->next = *walk;
    *walk = new;
}


extern unsigned long loops_per_jiffy;


void umlsim_setup_time(void)
{
    jiffie = 1.0/hz();
    jiffie_tv.tv_sec = 0;
    jiffie_tv.tv_usec = 1000000/hz();
    loops_per_jiffy = 1;
}
